package org.eclipse.jdt.internal.compiler.lookup;

import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.core.compiler.CharOperation;
import java.util.Set;
import java.util.List;
import org.eclipse.jdt.internal.compiler.util.Sorting;
import java.util.HashSet;
import java.util.ArrayList;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.util.SimpleSet;
import org.eclipse.jdt.internal.compiler.impl.CompilerOptions;
import org.eclipse.jdt.internal.compiler.ast.MethodDeclaration;
import org.eclipse.jdt.internal.compiler.ast.TypeDeclaration;
import org.eclipse.jdt.internal.compiler.util.HashtableOfObject;

public abstract class MethodVerifier extends ImplicitNullAnnotationVerifier
{
    SourceTypeBinding type;
    HashtableOfObject inheritedMethods;
    HashtableOfObject currentMethods;
    HashtableOfObject inheritedOverriddenMethods;
    
    MethodVerifier(LookupEnvironment environment) {
        super(environment);
        this.type = null;
        this.inheritedMethods = null;
        this.currentMethods = null;
        this.inheritedOverriddenMethods = null;
    }
    
    boolean areMethodsCompatible(MethodBinding one, MethodBinding two) {
        return areMethodsCompatible(one, two, this.environment);
    }
    
    static boolean areMethodsCompatible(MethodBinding one, MethodBinding two, LookupEnvironment environment) {
        one = one.original();
        two = one.findOriginalInheritedMethod(two);
        return two != null && isParameterSubsignature(one, two, environment);
    }
    
    boolean areReturnTypesCompatible(MethodBinding one, MethodBinding two) {
        return areReturnTypesCompatible(one, two, this.type.scope.environment());
    }
    
    public static boolean areReturnTypesCompatible(MethodBinding one, MethodBinding two, LookupEnvironment environment) {
        if (TypeBinding.equalsEquals(one.returnType, two.returnType)) {
            return true;
        }
        if (environment.globalOptions.sourceLevel >= 3211264L) {
            if (one.returnType.isBaseType()) {
                return false;
            }
            if (!one.declaringClass.isInterface() && one.declaringClass.id == 1) {
                return two.returnType.isCompatibleWith(one.returnType);
            }
            return one.returnType.isCompatibleWith(two.returnType);
        }
        else {
            return ImplicitNullAnnotationVerifier.areTypesEqual(one.returnType.erasure(), two.returnType.erasure());
        }
    }
    
    boolean canSkipInheritedMethods() {
        return (this.type.superclass() == null || !this.type.superclass().isAbstract()) && this.type.superInterfaces() == Binding.NO_SUPERINTERFACES;
    }
    
    boolean canSkipInheritedMethods(MethodBinding one, MethodBinding two) {
        return two == null || TypeBinding.equalsEquals(one.declaringClass, two.declaringClass);
    }
    
    void checkAbstractMethod(MethodBinding abstractMethod) {
        if (this.mustImplementAbstractMethod(abstractMethod.declaringClass)) {
            TypeDeclaration typeDeclaration = this.type.scope.referenceContext;
            if (typeDeclaration != null) {
                MethodDeclaration missingAbstractMethod = typeDeclaration.addMissingAbstractMethodFor(abstractMethod);
                missingAbstractMethod.scope.problemReporter().abstractMethodMustBeImplemented(this.type, abstractMethod);
            }
            else {
                this.problemReporter().abstractMethodMustBeImplemented(this.type, abstractMethod);
            }
        }
    }
    
    void checkAgainstInheritedMethods(MethodBinding currentMethod, MethodBinding[] methods, int length, MethodBinding[] allInheritedMethods) {
        if (this.type.isAnnotationType()) {
            this.problemReporter().annotationCannotOverrideMethod(currentMethod, methods[length - 1]);
            return;
        }
        CompilerOptions options = this.type.scope.compilerOptions();
        int[] overriddenInheritedMethods = (length > 1) ? this.findOverriddenInheritedMethods(methods, length) : null;
        int i = length;
    Label_0492:
        while (--i >= 0) {
            MethodBinding inheritedMethod = methods[i];
            if (overriddenInheritedMethods == null || overriddenInheritedMethods[i] == 0) {
                if (currentMethod.isStatic() != inheritedMethod.isStatic()) {
                    this.problemReporter(currentMethod).staticAndInstanceConflict(currentMethod, inheritedMethod);
                    continue;
                }
                if (inheritedMethod.isAbstract()) {
                    if (inheritedMethod.declaringClass.isInterface()) {
                        currentMethod.modifiers |= 0x20000000;
                    }
                    else {
                        currentMethod.modifiers |= 0x30000000;
                    }
                }
                else if (inheritedMethod.isPublic() || !this.type.isInterface()) {
                    if (currentMethod.isDefaultMethod() && !inheritedMethod.isFinal() && inheritedMethod.declaringClass.id == 1) {
                        this.problemReporter(currentMethod).defaultMethodOverridesObjectMethod(currentMethod);
                    }
                    else if (inheritedMethod.isDefaultMethod()) {
                        currentMethod.modifiers |= 0x20000000;
                    }
                    else {
                        currentMethod.modifiers |= 0x10000000;
                    }
                }
                if (!this.areReturnTypesCompatible(currentMethod, inheritedMethod) && (currentMethod.returnType.tagBits & 0x80L) == 0x0L && this.reportIncompatibleReturnTypeError(currentMethod, inheritedMethod)) {
                    continue;
                }
                this.reportRawReferences(currentMethod, inheritedMethod);
                if (currentMethod.thrownExceptions != Binding.NO_EXCEPTIONS) {
                    this.checkExceptions(currentMethod, inheritedMethod);
                }
                if (inheritedMethod.isFinal()) {
                    this.problemReporter(currentMethod).finalMethodCannotBeOverridden(currentMethod, inheritedMethod);
                }
                if (!this.isAsVisible(currentMethod, inheritedMethod)) {
                    this.problemReporter(currentMethod).visibilityConflict(currentMethod, inheritedMethod);
                }
                if (inheritedMethod.isSynchronized() && !currentMethod.isSynchronized()) {
                    this.problemReporter(currentMethod).missingSynchronizedOnInheritedMethod(currentMethod, inheritedMethod);
                }
                if (options.reportDeprecationWhenOverridingDeprecatedMethod && inheritedMethod.isViewedAsDeprecated() && (!currentMethod.isViewedAsDeprecated() || options.reportDeprecationInsideDeprecatedCode)) {
                    ReferenceBinding declaringClass = inheritedMethod.declaringClass;
                    if (declaringClass.isInterface()) {
                        int j = length;
                        while (--j >= 0) {
                            if (i != j && methods[j].declaringClass.implementsInterface(declaringClass, false)) {
                                continue Label_0492;
                            }
                            continue;
                        }
                    }
                    this.problemReporter(currentMethod).overridesDeprecatedMethod(currentMethod, inheritedMethod);
                }
            }
            if (!inheritedMethod.isStatic() && !inheritedMethod.isFinal()) {
                this.checkForBridgeMethod(currentMethod, inheritedMethod, allInheritedMethods);
            }
            else {
                continue;
            }
        }
        MethodBinding[] overridden = (MethodBinding[])this.inheritedOverriddenMethods.get(currentMethod.selector);
        if (overridden != null) {
            int k = overridden.length;
            while (--k >= 0) {
                MethodBinding inheritedMethod2 = overridden[k];
                if (this.isParameterSubsignature(currentMethod, inheritedMethod2) && !inheritedMethod2.isStatic() && !inheritedMethod2.isFinal()) {
                    this.checkForBridgeMethod(currentMethod, inheritedMethod2, allInheritedMethods);
                }
                else {
                    continue;
                }
            }
        }
        return;
    }
    
    void addBridgeMethodCandidate(MethodBinding overriddenMethod) {
        MethodBinding[] existing = (MethodBinding[])this.inheritedOverriddenMethods.get(overriddenMethod.selector);
        if (existing == null) {
            existing = new MethodBinding[] { overriddenMethod };
        }
        else {
            int length = existing.length;
            System.arraycopy(existing, 0, existing = new MethodBinding[length + 1], 0, length);
            existing[length] = overriddenMethod;
        }
        this.inheritedOverriddenMethods.put(overriddenMethod.selector, existing);
    }
    
    public void reportRawReferences(MethodBinding currentMethod, MethodBinding inheritedMethod) {
    }
    
    void checkConcreteInheritedMethod(MethodBinding concreteMethod, MethodBinding[] abstractMethods) {
        if (concreteMethod.isStatic()) {
            this.problemReporter().staticInheritedMethodConflicts(this.type, concreteMethod, abstractMethods);
        }
        if (!concreteMethod.isPublic()) {
            int index = 0;
            int length = abstractMethods.length;
            if (concreteMethod.isProtected()) {
                while (index < length) {
                    if (abstractMethods[index].isPublic()) {
                        break;
                    }
                    ++index;
                }
            }
            else if (concreteMethod.isDefault()) {
                while (index < length) {
                    if (!abstractMethods[index].isDefault()) {
                        break;
                    }
                    ++index;
                }
            }
            if (index < length) {
                this.problemReporter().inheritedMethodReducesVisibility(this.type, concreteMethod, abstractMethods);
            }
        }
        if (concreteMethod.thrownExceptions != Binding.NO_EXCEPTIONS) {
            int i = abstractMethods.length;
            while (--i >= 0) {
                this.checkExceptions(concreteMethod, abstractMethods[i]);
            }
        }
        if (concreteMethod.isOrEnclosedByPrivateType()) {
            MethodBinding original = concreteMethod.original();
            original.modifiers |= 0x8000000;
        }
    }
    
    void checkExceptions(MethodBinding newMethod, MethodBinding inheritedMethod) {
        ReferenceBinding[] newExceptions = this.resolvedExceptionTypesFor(newMethod);
        ReferenceBinding[] inheritedExceptions = this.resolvedExceptionTypesFor(inheritedMethod);
        int i = newExceptions.length;
        while (--i >= 0) {
            ReferenceBinding newException = newExceptions[i];
            int j = inheritedExceptions.length;
            while (--j > -1 && !this.isSameClassOrSubclassOf(newException, inheritedExceptions[j])) {}
            if (j == -1 && !newException.isUncheckedException(false) && (newException.tagBits & 0x80L) == 0x0L) {
                this.problemReporter(newMethod).incompatibleExceptionInThrowsClause(this.type, newMethod, inheritedMethod, newException);
            }
            else {
                continue;
            }
        }
    }
    
    void checkForBridgeMethod(MethodBinding currentMethod, MethodBinding inheritedMethod, MethodBinding[] allInheritedMethods) {
    }
    
    void checkForMissingHashCodeMethod() {
        MethodBinding[] choices = this.type.getMethods(TypeConstants.EQUALS);
        boolean overridesEquals = false;
        for (int i = choices.length; !overridesEquals && --i >= 0; overridesEquals = (choices[i].parameters.length == 1 && choices[i].parameters[0].id == 1)) {}
        if (overridesEquals) {
            MethodBinding hashCodeMethod = this.type.getExactMethod(TypeConstants.HASHCODE, Binding.NO_PARAMETERS, null);
            if (hashCodeMethod != null && hashCodeMethod.declaringClass.id == 1) {
                this.problemReporter().shouldImplementHashcode(this.type);
            }
        }
    }
    
    void checkForRedundantSuperinterfaces(ReferenceBinding superclass, ReferenceBinding[] superInterfaces) {
        if (superInterfaces == Binding.NO_SUPERINTERFACES) {
            return;
        }
        SimpleSet interfacesToCheck = new SimpleSet(superInterfaces.length);
        SimpleSet redundantInterfaces = null;
        for (int i = 0, l = superInterfaces.length; i < l; ++i) {
            ReferenceBinding toCheck = superInterfaces[i];
            for (int j = 0; j < l; ++j) {
                ReferenceBinding implementedInterface = superInterfaces[j];
                if (i != j && toCheck.implementsInterface(implementedInterface, true)) {
                    if (redundantInterfaces == null) {
                        redundantInterfaces = new SimpleSet(3);
                    }
                    else if (redundantInterfaces.includes(implementedInterface)) {
                        continue;
                    }
                    redundantInterfaces.add(implementedInterface);
                    TypeReference[] refs = this.type.scope.referenceContext.superInterfaces;
                    int r = 0;
                    int rl = refs.length;
                    while (r < rl) {
                        if (TypeBinding.equalsEquals(refs[r].resolvedType, toCheck)) {
                            this.problemReporter().redundantSuperInterface(this.type, refs[j], implementedInterface, toCheck);
                            break;
                        }
                        ++r;
                    }
                }
            }
            interfacesToCheck.add(toCheck);
        }
        ReferenceBinding[] itsInterfaces = null;
        SimpleSet inheritedInterfaces = new SimpleSet(5);
        for (ReferenceBinding superType = superclass; superType != null && superType.isValidBinding(); superType = superType.superclass()) {
            if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES) {
                for (int k = 0, m = itsInterfaces.length; k < m; ++k) {
                    ReferenceBinding inheritedInterface = itsInterfaces[k];
                    if (!inheritedInterfaces.includes(inheritedInterface) && inheritedInterface.isValidBinding()) {
                        if (interfacesToCheck.includes(inheritedInterface)) {
                            if (redundantInterfaces == null) {
                                redundantInterfaces = new SimpleSet(3);
                            }
                            else if (redundantInterfaces.includes(inheritedInterface)) {
                                continue;
                            }
                            redundantInterfaces.add(inheritedInterface);
                            TypeReference[] refs2 = this.type.scope.referenceContext.superInterfaces;
                            int r2 = 0;
                            int rl2 = refs2.length;
                            while (r2 < rl2) {
                                if (TypeBinding.equalsEquals(refs2[r2].resolvedType, inheritedInterface)) {
                                    this.problemReporter().redundantSuperInterface(this.type, refs2[r2], inheritedInterface, superType);
                                    break;
                                }
                                ++r2;
                            }
                        }
                        else {
                            inheritedInterfaces.add(inheritedInterface);
                        }
                    }
                }
            }
        }
        int nextPosition = inheritedInterfaces.elementSize;
        if (nextPosition == 0) {
            return;
        }
        ReferenceBinding[] interfacesToVisit = new ReferenceBinding[nextPosition];
        inheritedInterfaces.asArray(interfacesToVisit);
        for (int i2 = 0; i2 < nextPosition; ++i2) {
            ReferenceBinding superType = interfacesToVisit[i2];
            if ((itsInterfaces = superType.superInterfaces()) != Binding.NO_SUPERINTERFACES) {
                int itsLength = itsInterfaces.length;
                if (nextPosition + itsLength >= interfacesToVisit.length) {
                    System.arraycopy(interfacesToVisit, 0, interfacesToVisit = new ReferenceBinding[nextPosition + itsLength + 5], 0, nextPosition);
                }
                for (ReferenceBinding inheritedInterface2 : itsInterfaces) {
                    if (!inheritedInterfaces.includes(inheritedInterface2) && inheritedInterface2.isValidBinding()) {
                        if (interfacesToCheck.includes(inheritedInterface2)) {
                            if (redundantInterfaces == null) {
                                redundantInterfaces = new SimpleSet(3);
                            }
                            else if (redundantInterfaces.includes(inheritedInterface2)) {
                                continue;
                            }
                            redundantInterfaces.add(inheritedInterface2);
                            TypeReference[] refs3 = this.type.scope.referenceContext.superInterfaces;
                            int r3 = 0;
                            int rl3 = refs3.length;
                            while (r3 < rl3) {
                                if (TypeBinding.equalsEquals(refs3[r3].resolvedType, inheritedInterface2)) {
                                    this.problemReporter().redundantSuperInterface(this.type, refs3[r3], inheritedInterface2, superType);
                                    break;
                                }
                                ++r3;
                            }
                        }
                        else {
                            inheritedInterfaces.add(inheritedInterface2);
                            interfacesToVisit[nextPosition++] = inheritedInterface2;
                        }
                    }
                }
            }
        }
        return;
    }
    
    void checkInheritedMethods(MethodBinding[] methods, int length, boolean[] isOverridden, boolean[] isInherited) {
        MethodBinding concreteMethod = (this.type.isInterface() || methods[0].isAbstract()) ? null : methods[0];
        if (concreteMethod == null) {
            MethodBinding bestAbstractMethod2 = (length == 1) ? methods[0] : this.findBestInheritedAbstractOrDefaultMethod(methods, length);
            boolean noMatch = bestAbstractMethod2 == null;
            if (noMatch) {
                bestAbstractMethod2 = methods[0];
            }
            if (this.mustImplementAbstractMethod(bestAbstractMethod2.declaringClass)) {
                TypeDeclaration typeDeclaration = this.type.scope.referenceContext;
                MethodBinding superclassAbstractMethod = methods[0];
                if (superclassAbstractMethod == bestAbstractMethod2 || superclassAbstractMethod.declaringClass.isInterface()) {
                    if (typeDeclaration != null) {
                        MethodDeclaration missingAbstractMethod = typeDeclaration.addMissingAbstractMethodFor(bestAbstractMethod2);
                        missingAbstractMethod.scope.problemReporter().abstractMethodMustBeImplemented(this.type, bestAbstractMethod2);
                    }
                    else {
                        this.problemReporter().abstractMethodMustBeImplemented(this.type, bestAbstractMethod2);
                    }
                }
                else if (typeDeclaration != null) {
                    MethodDeclaration missingAbstractMethod = typeDeclaration.addMissingAbstractMethodFor(bestAbstractMethod2);
                    missingAbstractMethod.scope.problemReporter().abstractMethodMustBeImplemented(this.type, bestAbstractMethod2, superclassAbstractMethod);
                }
                else {
                    this.problemReporter().abstractMethodMustBeImplemented(this.type, bestAbstractMethod2, superclassAbstractMethod);
                }
            }
            else if (noMatch) {
                this.problemReporter().inheritedMethodsHaveIncompatibleReturnTypes(this.type, methods, length, isOverridden);
            }
            return;
        }
        else {
            if (length < 2) {
                return;
            }
            int index = length;
            while (--index > 0 && this.checkInheritedReturnTypes(concreteMethod, methods[index])) {}
            if (index > 0) {
                MethodBinding bestAbstractMethod = this.findBestInheritedAbstractOrDefaultMethod(methods, length);
                if (bestAbstractMethod == null) {
                    this.problemReporter().inheritedMethodsHaveIncompatibleReturnTypes(this.type, methods, length, isOverridden);
                }
                else {
                    this.problemReporter().abstractMethodMustBeImplemented(this.type, bestAbstractMethod, concreteMethod);
                }
                return;
            }
            else {
                MethodBinding[] abstractMethods = new MethodBinding[length - 1];
                index = 0;
                for (int i = 0; i < length; ++i) {
                    if (methods[i].isAbstract() || (methods[i] != concreteMethod && methods[i].isDefaultMethod())) {
                        abstractMethods[index++] = methods[i];
                    }
                }
                if (index == 0) {
                    return;
                }
                if (index < abstractMethods.length) {
                    System.arraycopy(abstractMethods, 0, abstractMethods = new MethodBinding[index], 0, index);
                }
                this.checkConcreteInheritedMethod(concreteMethod, abstractMethods);
                return;
            }
        }
    }
    
    boolean checkInheritedReturnTypes(MethodBinding method, MethodBinding otherMethod) {
        return this.areReturnTypesCompatible(method, otherMethod) || (!this.type.isInterface() && (method.declaringClass.isClass() || !this.type.implementsInterface(method.declaringClass, false)) && (otherMethod.declaringClass.isClass() || !this.type.implementsInterface(otherMethod.declaringClass, false)));
    }
    
    abstract void checkMethods();
    
    void checkPackagePrivateAbstractMethod(MethodBinding abstractMethod) {
        PackageBinding necessaryPackage = abstractMethod.declaringClass.fPackage;
        if (necessaryPackage == this.type.fPackage) {
            return;
        }
        ReferenceBinding superType = this.type.superclass();
        char[] selector = abstractMethod.selector;
        while (superType.isValidBinding()) {
            if (!superType.isAbstract()) {
                return;
            }
            if (necessaryPackage == superType.fPackage) {
                MethodBinding[] methods = superType.getMethods(selector);
                int m = methods.length;
                while (--m >= 0) {
                    MethodBinding method = methods[m];
                    if (!method.isPrivate() && !method.isConstructor()) {
                        if (method.isDefaultAbstract()) {
                            continue;
                        }
                        if (this.areMethodsCompatible(method, abstractMethod)) {
                            return;
                        }
                        continue;
                    }
                    else {
                        continue;
                    }
                }
            }
            if (TypeBinding.notEquals(superType = superType.superclass(), abstractMethod.declaringClass)) {
                continue;
            }
            this.problemReporter().abstractMethodCannotBeOverridden(this.type, abstractMethod);
            return;
        }
        return;
    }
    
    void computeInheritedMethods() {
        ReferenceBinding superclass = this.type.isInterface() ? this.type.scope.getJavaLangObject() : this.type.superclass();
        this.computeInheritedMethods(superclass, this.type.superInterfaces());
        this.checkForRedundantSuperinterfaces(superclass, this.type.superInterfaces());
    }
    
    void computeInheritedMethods(ReferenceBinding superclass, ReferenceBinding[] superInterfaces) {
        this.inheritedMethods = new HashtableOfObject(51);
        this.inheritedOverriddenMethods = new HashtableOfObject(11);
        ReferenceBinding superType = superclass;
        HashtableOfObject nonVisibleDefaultMethods = new HashtableOfObject(3);
        while (superType != null && superType.isValidBinding()) {
            MethodBinding[] methods = superType.unResolvedMethods();
            int m = methods.length;
        Label_0642:
            while (--m >= 0) {
                MethodBinding inheritedMethod = methods[m];
                if (!inheritedMethod.isPrivate() && !inheritedMethod.isConstructor()) {
                    if (inheritedMethod.isDefaultAbstract()) {
                        continue;
                    }
                    MethodBinding[] existingMethods = (MethodBinding[])this.inheritedMethods.get(inheritedMethod.selector);
                    if (existingMethods != null) {
                        for (int i = 0, length = existingMethods.length; i < length; ++i) {
                            MethodBinding existingMethod = existingMethods[i];
                            if (TypeBinding.notEquals(existingMethod.declaringClass, inheritedMethod.declaringClass) && this.areMethodsCompatible(existingMethod, inheritedMethod) && !this.canOverridingMethodDifferInErasure(existingMethod, inheritedMethod)) {
                                if (inheritedMethod.isDefault()) {
                                    if (inheritedMethod.isAbstract()) {
                                        this.checkPackagePrivateAbstractMethod(inheritedMethod);
                                    }
                                    else if (existingMethod.declaringClass.fPackage != inheritedMethod.declaringClass.fPackage && this.type.fPackage == inheritedMethod.declaringClass.fPackage && !this.areReturnTypesCompatible(inheritedMethod, existingMethod)) {
                                        continue;
                                    }
                                }
                                if (TypeBinding.notEquals(inheritedMethod.returnType.erasure(), existingMethod.returnType.erasure()) && this.areReturnTypesCompatible(existingMethod, inheritedMethod)) {
                                    this.addBridgeMethodCandidate(inheritedMethod);
                                    continue Label_0642;
                                }
                                continue Label_0642;
                            }
                        }
                    }
                    if (!inheritedMethod.isDefault() || inheritedMethod.declaringClass.fPackage == this.type.fPackage) {
                        if (existingMethods == null) {
                            existingMethods = new MethodBinding[] { inheritedMethod };
                        }
                        else {
                            int length3 = existingMethods.length;
                            System.arraycopy(existingMethods, 0, existingMethods = new MethodBinding[length3 + 1], 0, length3);
                            existingMethods[length3] = inheritedMethod;
                        }
                        this.inheritedMethods.put(inheritedMethod.selector, existingMethods);
                    }
                    else {
                        MethodBinding[] nonVisible = (MethodBinding[])nonVisibleDefaultMethods.get(inheritedMethod.selector);
                        if (nonVisible != null && inheritedMethod.isAbstract()) {
                            int j = 0;
                            int l = nonVisible.length;
                            while (j < l) {
                                if (this.areMethodsCompatible(nonVisible[j], inheritedMethod)) {
                                    continue Label_0642;
                                }
                                ++j;
                            }
                        }
                        if (nonVisible == null) {
                            nonVisible = new MethodBinding[] { inheritedMethod };
                        }
                        else {
                            int length = nonVisible.length;
                            System.arraycopy(nonVisible, 0, nonVisible = new MethodBinding[length + 1], 0, length);
                            nonVisible[length] = inheritedMethod;
                        }
                        nonVisibleDefaultMethods.put(inheritedMethod.selector, nonVisible);
                        if (inheritedMethod.isAbstract() && !this.type.isAbstract()) {
                            this.problemReporter().abstractMethodCannotBeOverridden(this.type, inheritedMethod);
                        }
                        MethodBinding[] current = (MethodBinding[])this.currentMethods.get(inheritedMethod.selector);
                        if (current != null && !inheritedMethod.isStatic()) {
                            int k = 0;
                            int length2 = current.length;
                            while (k < length2) {
                                if (!current[k].isStatic() && this.areMethodsCompatible(current[k], inheritedMethod)) {
                                    this.problemReporter().overridesPackageDefaultMethod(current[k], inheritedMethod);
                                    break;
                                }
                                ++k;
                            }
                        }
                        else {
                            continue;
                        }
                    }
                }
                else {
                    continue;
                }
            }
            superType = superType.superclass();
        }
        List<Object> superIfcList = new ArrayList();
        HashSet<Object> seenTypes = new HashSet();
        this.collectAllDistinctSuperInterfaces(superInterfaces, seenTypes, superIfcList);
        for (ReferenceBinding currentSuper = superclass; currentSuper != null && currentSuper.id != 1; currentSuper = currentSuper.superclass()) {
            this.collectAllDistinctSuperInterfaces(currentSuper.superInterfaces(), seenTypes, superIfcList);
        }
        if (superIfcList.size() == 0) {
            return;
        }
        if (superIfcList.size() == 1) {
            superInterfaces = new ReferenceBinding[] { (ReferenceBinding)superIfcList.get(0) };
        }
        else {
            superInterfaces = (ReferenceBinding[])superIfcList.toArray(new ReferenceBinding[superIfcList.size()]);
            superInterfaces = Sorting.sortTypes(superInterfaces);
        }
        SimpleSet skip = this.findSuperinterfaceCollisions(superclass, superInterfaces);
        int len = superInterfaces.length;
        for (int j = len - 1; j >= 0; --j) {
            superType = superInterfaces[j];
            if (superType.isValidBinding() && (skip == null || !skip.includes(superType))) {
                MethodBinding[] methods2 = superType.unResolvedMethods();
                int m2 = methods2.length;
            Label_1046:
                while (--m2 >= 0) {
                    MethodBinding inheritedMethod2 = methods2[m2];
                    if (inheritedMethod2.isStatic()) {
                        continue;
                    }
                    MethodBinding[] existingMethods2 = (MethodBinding[])this.inheritedMethods.get(inheritedMethod2.selector);
                    if (existingMethods2 == null) {
                        existingMethods2 = new MethodBinding[] { inheritedMethod2 };
                    }
                    else {
                        int length4 = existingMethods2.length;
                        for (int e = 0; e < length4; ++e) {
                            if (this.isInterfaceMethodImplemented(inheritedMethod2, existingMethods2[e], superType)) {
                                if (TypeBinding.notEquals(inheritedMethod2.returnType.erasure(), existingMethods2[e].returnType.erasure())) {
                                    this.addBridgeMethodCandidate(inheritedMethod2);
                                }
                                if (!this.canOverridingMethodDifferInErasure(existingMethods2[e], inheritedMethod2)) {
                                    continue Label_1046;
                                }
                            }
                        }
                        System.arraycopy(existingMethods2, 0, existingMethods2 = new MethodBinding[length4 + 1], 0, length4);
                        existingMethods2[length4] = inheritedMethod2;
                    }
                    this.inheritedMethods.put(inheritedMethod2.selector, existingMethods2);
                }
            }
        }
        return;
    }
    
    void collectAllDistinctSuperInterfaces(ReferenceBinding[] superInterfaces, Set seen, List result) {
        for (ReferenceBinding superInterface : superInterfaces) {
            if (seen.add(superInterface)) {
                result.add(superInterface);
                this.collectAllDistinctSuperInterfaces(superInterface.superInterfaces(), seen, result);
            }
        }
    }
    
    protected boolean canOverridingMethodDifferInErasure(MethodBinding overridingMethod, MethodBinding inheritedMethod) {
        return false;
    }
    
    void computeMethods() {
        MethodBinding[] methods = this.type.methods();
        int size = methods.length;
        this.currentMethods = new HashtableOfObject((size == 0) ? 1 : size);
        int m = size;
        while (--m >= 0) {
            MethodBinding method = methods[m];
            if (!method.isConstructor() && !method.isDefaultAbstract()) {
                MethodBinding[] existingMethods = (MethodBinding[])this.currentMethods.get(method.selector);
                if (existingMethods == null) {
                    existingMethods = new MethodBinding[] { null };
                }
                else {
                    System.arraycopy(existingMethods, 0, existingMethods = new MethodBinding[existingMethods.length + 1], 0, existingMethods.length - 1);
                }
                existingMethods[existingMethods.length - 1] = method;
                this.currentMethods.put(method.selector, existingMethods);
            }
            else {
                continue;
            }
        }
    }
    
    MethodBinding computeSubstituteMethod(MethodBinding inheritedMethod, MethodBinding currentMethod) {
        return computeSubstituteMethod(inheritedMethod, currentMethod, this.environment);
    }
    
    public static MethodBinding computeSubstituteMethod(MethodBinding inheritedMethod, MethodBinding currentMethod, LookupEnvironment environment) {
        if (inheritedMethod == null) {
            return null;
        }
        if (currentMethod.parameters.length != inheritedMethod.parameters.length) {
            return null;
        }
        if ((currentMethod.declaringClass instanceof BinaryTypeBinding)) {
            ((BinaryTypeBinding)currentMethod.declaringClass).resolveTypesFor(currentMethod);
        }
        if ((inheritedMethod.declaringClass instanceof BinaryTypeBinding)) {
            ((BinaryTypeBinding)inheritedMethod.declaringClass).resolveTypesFor(inheritedMethod);
        }
        TypeVariableBinding[] inheritedTypeVariables = inheritedMethod.typeVariables;
        int inheritedLength = inheritedTypeVariables.length;
        if (inheritedLength == 0) {
            return inheritedMethod;
        }
        TypeVariableBinding[] typeVariables = currentMethod.typeVariables;
        int length = typeVariables.length;
        if (length == 0) {
            return inheritedMethod.asRawMethod(environment);
        }
        if (length != inheritedLength) {
            return inheritedMethod;
        }
        TypeBinding[] arguments = new TypeBinding[length];
        System.arraycopy(typeVariables, 0, arguments, 0, length);
        ParameterizedGenericMethodBinding substitute = environment.createParameterizedGenericMethod(inheritedMethod, arguments);
        for (int i = 0; i < inheritedLength; ++i) {
            TypeVariableBinding inheritedTypeVariable = inheritedTypeVariables[i];
            TypeVariableBinding typeVariable = (TypeVariableBinding)arguments[i];
            if (TypeBinding.equalsEquals(typeVariable.firstBound, inheritedTypeVariable.firstBound)) {
                if (typeVariable.firstBound == null) {
                    continue;
                }
            }
            else if (typeVariable.firstBound != null && inheritedTypeVariable.firstBound != null && typeVariable.firstBound.isClass() != inheritedTypeVariable.firstBound.isClass()) {
                return inheritedMethod;
            }
            if (TypeBinding.notEquals(Scope.substitute(substitute, inheritedTypeVariable.superclass), typeVariable.superclass)) {
                return inheritedMethod;
            }
            int interfaceLength = inheritedTypeVariable.superInterfaces.length;
            ReferenceBinding[] interfaces = typeVariable.superInterfaces;
            if (interfaceLength != interfaces.length) {
                return inheritedMethod;
            }
            int j = 0;
        Label_0331:
            while (j < interfaceLength) {
                TypeBinding superType = Scope.substitute(substitute, inheritedTypeVariable.superInterfaces[j]);
                int k = 0;
                while (k < interfaceLength) {
                    if (TypeBinding.equalsEquals(superType, interfaces[k])) {
                        ++j;
                        continue Label_0331;
                    }
                    ++k;
                }
                return inheritedMethod;
            }
        }
        return substitute;
    }
    
    static boolean couldMethodOverride(MethodBinding method, MethodBinding inheritedMethod) {
        if (!CharOperation.equals(method.selector, inheritedMethod.selector)) {
            return false;
        }
        if (method == inheritedMethod || method.isStatic() || inheritedMethod.isStatic()) {
            return false;
        }
        if (inheritedMethod.isPrivate()) {
            return false;
        }
        if (inheritedMethod.isDefault() && method.declaringClass.getPackage() != inheritedMethod.declaringClass.getPackage()) {
            return false;
        }
        if (!method.isPublic()) {
            if (inheritedMethod.isPublic()) {
                return false;
            }
            if (inheritedMethod.isProtected() && !method.isProtected()) {
                return false;
            }
        }
        return true;
    }
    
    public boolean doesMethodOverride(MethodBinding method, MethodBinding inheritedMethod) {
        return doesMethodOverride(method, inheritedMethod, this.environment);
    }
    
    public static boolean doesMethodOverride(MethodBinding method, MethodBinding inheritedMethod, LookupEnvironment environment) {
        return couldMethodOverride(method, inheritedMethod) && areMethodsCompatible(method, inheritedMethod, environment);
    }
    
    SimpleSet findSuperinterfaceCollisions(ReferenceBinding superclass, ReferenceBinding[] superInterfaces) {
        return null;
    }
    
    MethodBinding findBestInheritedAbstractOrDefaultMethod(MethodBinding[] methods, int length) {
    Label_0096:
        for (int i = 0; i < length; ++i) {
            MethodBinding method = methods[i];
            if (method.isAbstract() || method.isDefaultMethod()) {
                for (int j = 0; j < length; ++j) {
                    if (i != j && !this.checkInheritedReturnTypes(method, methods[j])) {
                        if (this.type.isInterface() && methods[j].declaringClass.id == 1) {
                            return method;
                        }
                        continue Label_0096;
                    }
                }
                return method;
            }
        }
        return null;
    }
    
    int[] findOverriddenInheritedMethods(MethodBinding[] methods, int length) {
        int[] toSkip = null;
        int i = 0;
        ReferenceBinding declaringClass = methods[i].declaringClass;
        if (!declaringClass.isInterface()) {
            ReferenceBinding declaringClass2 = methods[++i].declaringClass;
            while (TypeBinding.equalsEquals(declaringClass, declaringClass2)) {
                if (++i == length) {
                    return null;
                }
                declaringClass2 = methods[i].declaringClass;
            }
            if (!declaringClass2.isInterface()) {
                if (declaringClass.fPackage != declaringClass2.fPackage && methods[i].isDefault()) {
                    return null;
                }
                toSkip = new int[length];
                do {
                    toSkip[i] = -1;
                    if (++i == length) {
                        return toSkip;
                    }
                    declaringClass2 = methods[i].declaringClass;
                } while (!declaringClass2.isInterface());
            }
        }
        while (i < length) {
            if (toSkip == null || toSkip[i] != -1) {
                declaringClass = methods[i].declaringClass;
                for (int j = i + 1; j < length; ++j) {
                    if (toSkip == null || toSkip[j] != -1) {
                        ReferenceBinding declaringClass3 = methods[j].declaringClass;
                        if (!TypeBinding.equalsEquals(declaringClass, declaringClass3)) {
                            if (declaringClass.implementsInterface(declaringClass3, true)) {
                                if (toSkip == null) {
                                    toSkip = new int[length];
                                }
                                toSkip[j] = -1;
                            }
                            else if (declaringClass3.implementsInterface(declaringClass, true)) {
                                if (toSkip == null) {
                                    toSkip = new int[length];
                                }
                                toSkip[i] = -1;
                                break;
                            }
                        }
                    }
                }
            }
            ++i;
        }
        return toSkip;
    }
    
    boolean isAsVisible(MethodBinding newMethod, MethodBinding inheritedMethod) {
        return inheritedMethod.modifiers == newMethod.modifiers || newMethod.isPublic() || (!inheritedMethod.isPublic() && (newMethod.isProtected() || (!inheritedMethod.isProtected() && !newMethod.isPrivate())));
    }
    
    boolean isInterfaceMethodImplemented(MethodBinding inheritedMethod, MethodBinding existingMethod, ReferenceBinding superType) {
        return ImplicitNullAnnotationVerifier.areParametersEqual(existingMethod, inheritedMethod) && existingMethod.declaringClass.implementsInterface(superType, true);
    }
    
    public boolean isMethodSubsignature(MethodBinding method, MethodBinding inheritedMethod) {
        return CharOperation.equals(method.selector, inheritedMethod.selector) && this.isParameterSubsignature(method, inheritedMethod);
    }
    
    boolean isParameterSubsignature(MethodBinding method, MethodBinding inheritedMethod) {
        return isParameterSubsignature(method, inheritedMethod, this.environment);
    }
    
    static boolean isParameterSubsignature(MethodBinding method, MethodBinding inheritedMethod, LookupEnvironment environment) {
        MethodBinding substitute = computeSubstituteMethod(inheritedMethod, method, environment);
        return substitute != null && isSubstituteParameterSubsignature(method, substitute, environment);
    }
    
    boolean isSubstituteParameterSubsignature(MethodBinding method, MethodBinding substituteMethod) {
        return isSubstituteParameterSubsignature(method, substituteMethod, this.environment);
    }
    
    public static boolean isSubstituteParameterSubsignature(MethodBinding method, MethodBinding substituteMethod, LookupEnvironment environment) {
        if (!ImplicitNullAnnotationVerifier.areParametersEqual(method, substituteMethod)) {
            if (substituteMethod.hasSubstitutedParameters() && method.areParameterErasuresEqual(substituteMethod)) {
                return method.typeVariables == Binding.NO_TYPE_VARIABLES && !hasGenericParameter(method);
            }
            return method.declaringClass.isRawType() && substituteMethod.declaringClass.isRawType() && method.hasSubstitutedParameters() && substituteMethod.hasSubstitutedParameters() && areMethodsCompatible(method, substituteMethod, environment);
        }
        else if ((substituteMethod instanceof ParameterizedGenericMethodBinding)) {
            if (method.typeVariables != Binding.NO_TYPE_VARIABLES) {
                return !((ParameterizedGenericMethodBinding)substituteMethod).isRaw;
            }
            return !hasGenericParameter(method);
        }
        else {
            return method.typeVariables == Binding.NO_TYPE_VARIABLES;
        }
    }
    
    static boolean hasGenericParameter(MethodBinding method) {
        if (method.genericSignature() == null) {
            return false;
        }
        TypeBinding[] params = method.parameters;
        for (int i = 0, l = params.length; i < l; ++i) {
            TypeBinding param = params[i].leafComponentType();
            if ((param instanceof ReferenceBinding)) {
                int modifiers = ((ReferenceBinding)param).modifiers;
                if ((modifiers & 0x40000000) != 0x0) {
                    return true;
                }
            }
        }
        return false;
    }
    
    boolean isSameClassOrSubclassOf(ReferenceBinding testClass, ReferenceBinding superclass) {
        while (!TypeBinding.equalsEquals(testClass, superclass)) {
            if ((testClass = testClass.superclass()) != null) {
                continue;
            }
            return false;
        }
        return true;
    }
    
    boolean mustImplementAbstractMethod(ReferenceBinding declaringClass) {
        if (!this.mustImplementAbstractMethods()) {
            return false;
        }
        ReferenceBinding superclass = this.type.superclass();
        if (declaringClass.isClass()) {
            while (superclass.isAbstract()) {
                if (TypeBinding.notEquals(superclass, declaringClass)) {
                    superclass = superclass.superclass();
                    continue;
                }
                break;
            }
        }
        else {
            if (this.type.implementsInterface(declaringClass, false) && !superclass.implementsInterface(declaringClass, true)) {
                return true;
            }
            while (superclass.isAbstract() && !superclass.implementsInterface(declaringClass, false)) {
                superclass = superclass.superclass();
            }
        }
        return superclass.isAbstract();
    }
    
    boolean mustImplementAbstractMethods() {
        return !this.type.isInterface() && !this.type.isAbstract();
    }
    
    ProblemReporter problemReporter() {
        return this.type.scope.problemReporter();
    }
    
    ProblemReporter problemReporter(MethodBinding currentMethod) {
        ProblemReporter reporter = this.problemReporter();
        if (TypeBinding.equalsEquals(currentMethod.declaringClass, this.type) && currentMethod.sourceMethod() != null) {
            reporter.referenceContext = currentMethod.sourceMethod();
            return reporter;
        }
        return reporter;
    }
    
    boolean reportIncompatibleReturnTypeError(MethodBinding currentMethod, MethodBinding inheritedMethod) {
        this.problemReporter(currentMethod).incompatibleReturnType(currentMethod, inheritedMethod);
        return true;
    }
    
    ReferenceBinding[] resolvedExceptionTypesFor(MethodBinding method) {
        ReferenceBinding[] exceptions = method.thrownExceptions;
        if ((method.modifiers & 0x2000000) == 0x0) {
            return exceptions;
        }
        if (!(method.declaringClass instanceof BinaryTypeBinding)) {
            return Binding.NO_EXCEPTIONS;
        }
        int i = exceptions.length;
        while (--i >= 0) {
            exceptions[i] = (ReferenceBinding)BinaryTypeBinding.resolveType(exceptions[i], this.environment, true);
        }
        return exceptions;
    }
    
    void verify() {
        this.computeMethods();
        this.computeInheritedMethods();
        this.checkMethods();
        if (this.type.isClass()) {
            this.checkForMissingHashCodeMethod();
        }
    }
    
    void verify(SourceTypeBinding someType) {
        if (this.type == null) {
            try {
                this.type = someType;
                this.verify();
            }
            finally {
                this.type = null;
            }
        }
        else {
            this.environment.newMethodVerifier().verify(someType);
        }
    }
    
    @Override
    public String toString() {
        StringBuffer buffer = new StringBuffer(10);
        buffer.append("MethodVerifier for type: ");
        buffer.append(this.type.readableName());
        buffer.append('\n');
        buffer.append("\t-inherited methods: ");
        buffer.append(this.inheritedMethods);
        return buffer.toString();
    }
}
